      .586
      .model flat, stdcall
      option casemap :none

      include \masm32\include\windows.inc
      include \masm32\include\user32.inc
      include \masm32\include\kernel32.inc
      include \masm32\include\masm32.inc
      include \masm32\include\fpu.inc
      include \masm32\macros\macros.asm
      includelib \masm32\lib\masm32.lib
      includelib \masm32\lib\user32.lib
      includelib \masm32\lib\kernel32.lib
      includelib \masm32\lib\fpu.lib

      ; Define for progress indicator.
      PROGRESS EQU 1
      
      ; Count of bytes to be generated.      
      COUNT    EQU 2048
      
      ; Use a tolerance of 13 to ensure a "and randomly
      ; would exceed this value 50.00 percent of the
      ; times" result, or 20 to ensure a result between
      ; 25% and 75%.
      CHISQTOL EQU 13

.data
      buffer      db 30 dup(0)
      randbuffer  db COUNT dup(0)
      filename    db "randbyte.bin",0
      startTC     dd 0
      csq         dd 0
      passes      dd 0
      bytecnt     dd 0
.code
start: 
      call GetTickCount
      mov   startTC,eax       
again:
      IFDEF PROGRESS      
            mov   bytecnt,0
            print SADD(13,10,"pass progress:")
            call getcursorx
            push  eax
            call getcursory
            push  eax
            mov   ebx,COUNT
            shr   ebx,8
         @@:
            print SADD("_")            
            dec   ebx
            jnz   @B
            pop   edx
            pop   eax
            invoke locate,eax,edx
      ENDIF

      mov   ebx,COUNT
      call GetTickCount
      mov   edi,eax
      rdtsc
      mov   esi,eax
      ; Avoid diluting randomness with unnecessary delays.
   runLoop:
      call GetTickCount
      mov   ecx,eax
      rdtsc
      cmp   ecx,edi
      je    runLoop
      mov   edi,ecx
      mov   ecx,eax
      sub   ecx,esi
      mov   esi,eax
      mov   eax,ecx

      ;SHR   EAX,3

      mov   [randbuffer+ebx-1],al

      IFDEF PROGRESS
            inc   bytecnt
            cmp   bytecnt,256
            jb    @F
            mov   bytecnt,0
            print SADD("=")
         @@:
      ENDIF

      dec   ebx
      jnz   runLoop

      IFDEF CHISQTOL
         ; Iterate until chi-square within tolerance.
         inc   passes
         call chisq
         fistp csq
         cmp   csq,256-CHISQTOL
         jb    again
         cmp   csq,256+CHISQTOL
         ja    again
      ENDIF

      invoke write_disk_file,addr filename,addr randbuffer,COUNT

      invoke dwtoa,passes,addr buffer
      print SADD(13,10,"passes = ")
      print addr buffer
      print SADD(13,10)

      call chisq      
      invoke FpuFLtoA,0,2,addr buffer,SRC1_FPU or SRC2_DIMM
      print SADD("chisq = ")
      print addr buffer
      print SADD(13,10)

      call GetTickCount
      sub   eax,startTC
      mov   ecx,1000
      div   ecx

      push  eax
      invoke dwtoa,eax,addr buffer
      print SADD("seconds = ")
      print addr buffer
      print SADD(13,10)
      pop   ecx

      mov   eax,COUNT
      xor   edx,edx
      div   ecx
      invoke dwtoa,eax,addr buffer
      print SADD("bytes/second = ")
      print addr buffer
      print SADD(13,10)

      print SADD(13,10,"Press enter to exit...",13,10)
      invoke StdIn,addr buffer,1
      exit

; 
; This procedure calculates the chi-square statistic
; for randbuffer and leaves the result on the FPU.
;
; Adapted from a Pascal example in Robert Sedgewick's
; Algorithms, Second Edition.
;
; Presumably, a "good" RNG will have a statistic that is
; within 2 * SQR(range) of the range (for N > 10 * range).
; 
chisq proc

      LOCAL cnt:DWORD,range:DWORD,sumsq:QWORD,freq[256]:DWORD
      
      mov   cnt,COUNT
      mov   range,256
      mov   dword ptr[sumsq],0
      mov   dword ptr[sumsq+4],0
      
      ; Init frequency array.
      mov   ecx,255
   @@:
      mov   dword ptr[freq+ecx*4],0
      dec   ecx
      jns   @B
      
      ; Fill frequency array.
      mov   ecx,cnt
      xor   eax,eax
   @@:
      mov   al,[randbuffer+ecx-1]
      inc   dword ptr[freq+eax*4]
      dec   ecx
      jnz   @B
      
      ; Get the sum of the squares of the frequencies.
      mov   ecx,255
   @@:
      mov   eax,[freq+ecx*4]
      mov   edx,eax
      mul   edx
      add   dword ptr[sumsq],eax
      adc   dword ptr[sumsq+4],edx
      dec   ecx
      jns   @B
      
      ; Calc chi-square.
      fild  range
      fild  sumsq    ; ST=sumsq,ST(1)=range
      fmul           ; ST=range*sumsq
      fild  cnt      ; ST=cnt,ST(1)=range*sumsq
      fdiv           ; ST=range*sumsq/cnt
      fild  cnt      ; ST=cnt,ST(1)=range*sumsq/cnt
      fsub           ; ST=(range*sumsq/cnt)-cnt
      ret
chisq endp
; 
getcursorx proc
      LOCAL hStdOut:DWORD
      LOCAL csbInfo:CONSOLE_SCREEN_BUFFER_INFO
      invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov	hStdOut,eax
	cmp	eax,INVALID_HANDLE_VALUE
	jne 	@F
      ret
   @@:
      invoke GetConsoleScreenBufferInfo,hStdOut,addr csbInfo
      movzx	eax,csbInfo.dwCursorPosition.x
      ret      
getcursorx endp

getcursory proc
      LOCAL hStdOut:DWORD
      LOCAL csbInfo:CONSOLE_SCREEN_BUFFER_INFO
      invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov	hStdOut,eax
	cmp	eax,INVALID_HANDLE_VALUE
	jne 	@F
      ret
   @@:
      invoke GetConsoleScreenBufferInfo,hStdOut,addr csbInfo
      movzx	eax,csbInfo.dwCursorPosition.y
      ret   
getcursory endp

end start
